#![no_std]
#![no_main]
#![allow(async_fn_in_trait)]
#![feature(slice_split_once)]

use core::fmt::{Debug, Write};
use core::str::from_utf8;
use cyw43_pio::{DEFAULT_CLOCK_DIVIDER, PioSpi};
use defmt::println;
use embassy_executor::Spawner;
use embassy_net::tcp::TcpSocket;
use embassy_net::{Config, StackResources};
use embassy_rp::bind_interrupts;
use embassy_rp::clocks::RoscRng;
use embassy_rp::gpio::{Level, Output};
use embassy_rp::peripherals::USB;
use embassy_rp::peripherals::{DMA_CH0, PIO0};
use embassy_rp::pio::{InterruptHandler as PioInterruptHandler, Pio};
use embassy_rp::usb::{Driver, InterruptHandler as UsbInterruptHandler};
use embassy_time::Duration;
use embassy_time::Timer;
use embedded_io_async::Write as _;
use heapless::{String, Vec};
use log::{info, warn};
use rand_core::RngCore;
use static_cell::StaticCell;
use {defmt_rtt as _, panic_probe as _};

bind_interrupts!(struct Irqs {
    USBCTRL_IRQ => UsbInterruptHandler<USB>;
    PIO0_IRQ_0 => PioInterruptHandler<PIO0>;
});

#[embassy_executor::task]
async fn logger_task(driver: Driver<'static, USB>) {
    embassy_usb_logger::run!(1024, log::LevelFilter::Info, driver);
}

#[embassy_executor::task]
async fn cyw43_task(
    runner: cyw43::Runner<'static, Output<'static>, PioSpi<'static, PIO0, 0, DMA_CH0>>,
) -> ! {
    runner.run().await
}

#[embassy_executor::task]
async fn net_task(mut runner: embassy_net::Runner<'static, cyw43::NetDriver<'static>>) -> ! {
    runner.run().await
}

#[embassy_executor::main]
async fn main(spawner: Spawner) {
    let p = embassy_rp::init(Default::default());
    let driver = Driver::new(p.USB, Irqs);
    spawner.spawn(logger_task(driver)).unwrap();
    let mut rng = RoscRng;

    let fw = include_bytes!("../cyw43-firmware/43439A0.bin");
    let clm = include_bytes!("../cyw43-firmware/43439A0_clm.bin");
    let pwr = Output::new(p.PIN_23, Level::Low);
    let cs = Output::new(p.PIN_25, Level::High);
    let mut pio = Pio::new(p.PIO0, Irqs);
    let spi = PioSpi::new(
        &mut pio.common,
        pio.sm0,
        DEFAULT_CLOCK_DIVIDER,
        pio.irq0,
        cs,
        p.PIN_24,
        p.PIN_29,
        p.DMA_CH0,
    );

    static STATE: StaticCell<cyw43::State> = StaticCell::new();
    let state = STATE.init(cyw43::State::new());
    let (net_device, mut control, runner) = cyw43::new(state, pwr, spi, fw).await;
    spawner.spawn(cyw43_task(runner)).unwrap();

    control.init(clm).await;
    control
        .set_power_management(cyw43::PowerManagementMode::PowerSave)
        .await;

    // Use a link-local address for communication without DHCP server
    let config = Config::ipv4_static(embassy_net::StaticConfigV4 {
        address: embassy_net::Ipv4Cidr::new(embassy_net::Ipv4Address::new(192, 254, 0, 2), 16),
        dns_servers: heapless::Vec::new(),
        gateway: None,
    });

    // Generate random seed
    let seed = rng.next_u64();

    // Init network stack
    static RESOURCES: StaticCell<StackResources<3>> = StaticCell::new();
    let (stack, runner) = embassy_net::new(
        net_device,
        config,
        RESOURCES.init(StackResources::new()),
        seed,
    );

    spawner.spawn(net_task(runner)).unwrap();

    //control.start_ap_open("cyw43", 5).await;
    control.start_ap_wpa2("cyw43", "password", 5).await;

    let mut rx_buffer = [0; 4096];
    let mut tx_buffer = [0; 4096];
    let mut buf = [0; 4096];
    let mut res_buf = Vec::<u8, 4096>::new();

    loop {
        let mut socket = TcpSocket::new(stack, &mut rx_buffer, &mut tx_buffer);
        socket.set_timeout(Some(Duration::from_secs(30)));

        control.gpio_set(0, false).await;
        info!("Listening on TCP:1234...");
        if let Err(e) = socket.accept(1234).await {
            warn!("accept error: {:?}", e);
            continue;
        }

        info!("Received connection from {:?}", socket.remote_endpoint());
        control.gpio_set(0, true).await;

        loop {
            let n = match socket.read(&mut buf).await {
                Ok(0) => {
                    warn!("read EOF");
                    break;
                }
                Ok(n) => n,
                Err(e) => {
                    warn!("read error: {:?}", e);
                    break;
                }
            };

            info!("request :\n{}", from_utf8(&buf[..n]).unwrap());

            let mut headers: &[u8] = &buf[..n];
            let mut content: &[u8] = &[];
            for i in 0..(n - 1) {
                if &buf[i..i + 1] == b"\r\n" {
                    headers = &buf[0..i];
                    if i + 2 < n {
                        content = &buf[i + 2..n];
                    }
                }
            }

            let mut headers = headers.split(|x| *x == b'\n');
            let (request_type, path) = match headers.next() {
                None => {
                    warn!("Empty request");
                    break;
                }
                Some(l1) => {
                    let mut l1 = l1.split(|x| *x == b' ');
                    (
                        match l1.next() {
                            Some(b"GET") => HttpRequestType::Get,
                            Some(b"POST") => HttpRequestType::Post,
                            Some(t) => {
                                warn!("Unknown request type : {}", from_utf8(t).unwrap());
                                break;
                            }
                            None => {
                                warn!("No request type");
                                break;
                            }
                        },
                        match l1.next() {
                            Some(path) => path,
                            None => {
                                warn!("No path");
                                break;
                            }
                        },
                    )
                }
            };

            let (code, res): (HttpResCode, &str) = match path {
                b"/" => (
                    HttpResCode::Ok,
                    "<!DOCTYPE html>\r\n\
                    <html>\r\n\
                    <body>\r\n\
                    <h1>Titre</h1>\r\n\
                    <p>contenu</p>\r\n\
                    </body>\r\n\
                    </html>",
                ),
                _ => (HttpResCode::NotFound, ""),
            };

            //b"HTTP/1.1 200 OK\r\n\
            // Content-Type: text/html\r\n\
            // Content-Length: 81\r\n\r\n\
            // <!DOCTYPE html>\r\n\
            // <html>\r\n\
            // <body>\r\n\
            // <h1>Titre</h1>\r\n\
            // <p>contenu</p>\r\n\
            // </body>\r\n\
            // </html>"

            // Write response to buf
            res_buf.clear();
            if let Err(e) = write!(
                &mut res_buf,
                "{}\r\n\
                Content-Type: text/html\r\n\
                Content-Length: {}\r\n\r\n\
                {}",
                Into::<&str>::into(code),
                res.len(),
                res
            ) {
                warn!("write error: {:?}", e);
                break;
            }

            info!("response :\n{}", from_utf8(&res_buf).unwrap());

            match socket.write_all(&res_buf).await {
                Ok(()) => {}
                Err(e) => {
                    warn!("write error: {:?}", e);
                    break;
                }
            };
        }
    }
}

enum HttpRequestType {
    Get,
    Post,
}

#[derive(Debug, Clone, Copy)]
enum HttpResCode {
    Ok,
    NotFound,
}
impl Into<&str> for HttpResCode {
    fn into(self) -> &'static str {
        match self {
            HttpResCode::Ok => "HTTP/1.1 200 OK",
            HttpResCode::NotFound => "HTTP/1.1 404 NOT FOUND",
        }
    }
}
